/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.model;

import se.llbit.chunky.world.Texture;
import se.llbit.math.AABB;
import se.llbit.math.QuickMath;
import se.llbit.math.Ray;

/**
 * A water block. The height of the top water block is slightly
 * lower than a regular block.
 * @author Jesper Öqvist <jesper@llbit.se>
 */
public class WaterModel {
	private static AABB fullBlock = new AABB(0, 1, 0, 1, 0, 1);
	private static AABB[] block = {
		new AABB(0, 1, 0, .875, 0, 1),
		new AABB(0, 1, 0, 14/16., 0, 1),
		new AABB(0, 1, 0, 12/16., 0, 1),
		new AABB(0, 1, 0, 10/16., 0, 1),
		new AABB(0, 1, 0, 8/16., 0, 1),
		new AABB(0, 1, 0, 6/16., 0, 1),
		new AABB(0, 1, 0, 4/16., 0, 1),
		new AABB(0, 1, 0, 2/16., 0, 1),
	};
	private static double opacity = .3;
	public static boolean stillWater = false;
	private static final float[][][] normalMap;
	private static final int normalMapW;

	static {
		// precompute normal map
		Texture waterHeight = new Texture("water-height");
		normalMapW = waterHeight.getWidth();
		normalMap = new float[normalMapW][normalMapW][2];
		for (int u = 0; u < normalMapW; ++u) {
			for (int v = 0; v < normalMapW; ++v) {
				
				float hx0 = (waterHeight.getColorWrapped(u, v) & 0xFF) / 255.f;
				float hx1 = (waterHeight.getColorWrapped(u+1, v) & 0xFF) / 255.f;
				float hz0 = (waterHeight.getColorWrapped(u, v) & 0xFF) / 255.f;
				float hz1 = (waterHeight.getColorWrapped(u, v+1) & 0xFF) / 255.f;
				normalMap[u][v][0] = hx1-hx0;
				normalMap[u][v][1] = hz1-hz0;
			}
		}
	}

	public static boolean intersect(Ray ray) {
		ray.t = Double.POSITIVE_INFINITY;
		
		int waterAbove = (ray.currentMaterial >> 16) & 1;
		int level = ray.currentMaterial >> 8;
		
		if (waterAbove != 0) {
			if ((level&8) != 0) {
				if (fullBlock.intersect(ray)) {
					if (ray.n.y != 0)
						return false;
					Texture.water.getAvgColorLinear(ray.color);
					ray.color.w = opacity;
					ray.distance += ray.tNear;
					ray.x.scaleAdd(ray.tNear, ray.d, ray.x);
					if (ray.n.x != 0) {
						double sign = Math.signum(ray.n.x);
						doWaterDisplacement(ray);
						ray.n.set(ray.n.y*sign, ray.n.x, ray.n.z);
					} else {
						double sign = Math.signum(ray.n.z);
						doWaterDisplacement(ray);
						ray.n.set(ray.n.x, ray.n.z, ray.n.y*sign);
					}
					return true;
				}
				return false;
			}
			if (fullBlock.intersect(ray)) {
				Texture.water.getAvgColorLinear(ray.color);
				ray.color.w = opacity;
				ray.distance += ray.tNear;
				ray.x.scaleAdd(ray.tNear, ray.d, ray.x);
				return true;
			}
			return false;
		}
		
		if (((ray.currentMaterial >> 17) & 1) != 0) {
			if (LilyPadModel.intersect(ray))
				return true;
		}
		if (block[level%8].intersect(ray)) {
			Texture.water.getAvgColorLinear(ray.color);
			ray.color.w = opacity;
			ray.distance += ray.tNear;
			ray.x.scaleAdd(ray.tNear, ray.d, ray.x);
			
			// Water normal displacement
			if (waterAbove == 0 && !stillWater && ray.n.y > 0) {
				doWaterDisplacement(ray);
			}
			
			return true;
		}
		return false;
	}

	private final static void doWaterDisplacement(Ray ray) {
		int w = (1<<4);
		double x = ray.x.x/w - QuickMath.floor(ray.x.x/w);
		double z = ray.x.z/w - QuickMath.floor(ray.x.z/w);
		int u = (int) (x * normalMapW - Ray.EPSILON);
		int v = (int) ((1-z) * normalMapW - Ray.EPSILON);
		ray.n.set(normalMap[u][v][0], .15f,
				normalMap[u][v][1]);
		w = (1<<1);
		x = ray.x.x/w - QuickMath.floor(ray.x.x/w);
		z = ray.x.z/w - QuickMath.floor(ray.x.z/w);
		u = (int) (x * normalMapW - Ray.EPSILON);
		v = (int) ((1-z) * normalMapW - Ray.EPSILON);
		ray.n.x += normalMap[u][v][0]/2;
		ray.n.z += normalMap[u][v][1]/2;
		ray.n.normalize();
	}
}
