/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.renderer;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.logging.Level;

import javax.imageio.ImageIO;

import se.llbit.chunky.main.Chunky;
import se.llbit.chunky.world.Texture;
import se.llbit.math.Color;
import se.llbit.math.QuickMath;
import se.llbit.math.Ray;
import se.llbit.nbt.AnyTag;
import se.llbit.nbt.CompoundTag;
import se.llbit.nbt.DoubleTag;
import se.llbit.nbt.StringTag;
import se.llbit.util.ProgramProperties;

public class Sky {
	private Texture skymap = null;
	private String skymapFileName = "";
	private Scene scene;
	private double rotation = 0;
	
	public Sky(Scene scene) {
		this.scene = scene;
	}
	
	public void loadSkyMap(String fileName) {
		skymapFileName = fileName;
		File sky = new File(skymapFileName);
		if (sky.exists()) {
			try {
				Chunky.logger.info("Loading sky map: " + fileName);
				skymap = new Texture(ImageIO.read(sky));
				ProgramProperties.setProperty("skymap", fileName);
			} catch (IOException e) {
				Chunky.logger.warning("Could not load skymap: " + fileName);
			} catch (Throwable e) {
				Chunky.logger.log(Level.SEVERE, "Unexpected exception ocurred!", e);
			}
		} else {
			Chunky.logger.warning("Skymap could not be opened: " + fileName);
		}
		scene.refresh();
	}
	
	public void set(Sky other) {
		skymapFileName = other.skymapFileName;
		skymap = other.skymap;
		rotation = other.rotation;
	}
	
	public synchronized void unloadSkymap() {
		skymapFileName = "";
		skymap = null;
		ProgramProperties.removeProperty("skymap");
		scene.refresh();
	}
	
	@SuppressWarnings("unused")
	private static final int SKYBOX = 1;
	private static BufferedImage top = null;//ImageLoader.get("top"+SKYBOX+".jpg");
	private static BufferedImage left = null;//ImageLoader.get("left"+SKYBOX+".jpg");
	private static BufferedImage right = null;//ImageLoader.get("right"+SKYBOX+".jpg");
	private static BufferedImage front = null;//ImageLoader.get("front"+SKYBOX+".jpg");
	private static BufferedImage back = null;//ImageLoader.get("back"+SKYBOX+".jpg");
	
	public void getSkyboxColor(Ray ray) {
		int face = 0;
		double tNear = Double.POSITIVE_INFINITY;
		double t1;
		double t2;
		if (ray.d.x != 0) {
			t1 = 1 / ray.d.x;
			t2 = Math.abs(t1);
			if (t2 < tNear) {
				tNear = t1;
				if (t1 > 0) {
					tNear = t1;
					face = 0;
				} else {
					tNear = t2;
					face = 1;
				}
			}
		}
		if (ray.d.y != 0) {
			t1 = Math.abs(1 / ray.d.y);
			if (t1 < tNear) {
				tNear = t1;
				face = 2;
			}
		}
		if (ray.d.z != 0) {
			t1 = 1 / ray.d.z;
			t2 = Math.abs(t1);
			if (t2 < tNear) {
				if (t1 > 0) {
					tNear = t1;
					face = 3;
				} else {
					tNear = t2;
					face = 4;
				}
			}
		}

		double x = ray.d.x * tNear;
		double y = ray.d.y * tNear;
		double z = ray.d.z * tNear;

		double u = 0, v = 0;
		BufferedImage texture = left;
		switch (face) {
			case 0:
				texture = right;
				v = 1 - (y / 2 + .5f);
				u = 1 - (z / 2 + .5f);
				break;
			case 1:
				texture = left;
				v = 1 - (y / 2 + .5f);
				u = z / 2 + .5f;
				break;
			case 2:
				texture = top;
				u = x / 2 + .5f;
				v = z / 2 + .5f;
				break;
			case 3:
				texture = front;
				u = x / 2 + .5f;
				v = 1 - (y / 2 + .5f);
				break;
			case 4:
				texture = back;
				u = 1 - (x / 2 + .5f);
				v = 1 - (y / 2 + .5f);
				break;
		}
		int rgb = texture.getRGB((int) (u * (texture.getWidth() - 1) + .5f),
				(int) (v * (texture.getHeight() - 1) + .5f));
		Color.getRGBComponents(rgb, ray.color);
		ray.hit = true;
	}

	/**
	 * Panormaic skymap color
	 * @param ray
	 * @param blackBelowHorizon
	 */
	public void getSkyPanoramaColor(Ray ray, boolean blackBelowHorizon) {
		if (blackBelowHorizon && ray.d.y < 0) {
			ray.color.set(0, 0, 0, 1);
			ray.hit = true;
			return;
		} else if (skymap == null) {
			scene.sun().skylight(ray);
			ray.hit = true;
			return;
		}
		double r = ray.d.z * ray.d.z + ray.d.x * ray.d.x;
		double theta = 0;
		if (r > Ray.EPSILON)
			theta = Math.asin(ray.d.z / Math.sqrt(r));
		if (ray.d.x < 0)
			theta = Math.PI - theta;
		theta += rotation;
		if (theta > 2 * Math.PI || theta < 0) {
			theta = theta % (2 * Math.PI);
			if (theta < 0)
				theta += 2 * Math.PI;
		}
		double phi = Math.abs(Math.asin(ray.d.y));
		skymap.getColor(theta / (2*Math.PI), (2 * phi / Math.PI), ray.color);
		ray.hit = true;
	}
	
	/**
	 * Bilinear interpolated panoramic skymap color
	 * @param ray
	 * @param blackBelowHorizon
	 */
	public void getSkyPanoramaColorInterpolated(Ray ray, boolean blackBelowHorizon) {
		if (blackBelowHorizon && ray.d.y < 0) {
			ray.color.set(0, 0, 0, 1);
			
		} else if (scene.sun().intersect(ray)) {
			double r = ray.color.x;
			double g = ray.color.y;
			double b = ray.color.z;
			getPanoramaColorInterpolated(ray);
			ray.color.x = ray.color.x + r;
			ray.color.y = ray.color.y + g;
			ray.color.z = ray.color.z + b;
			
		} else {
			
			getPanoramaColorInterpolated(ray);
		}
		ray.hit = true;
	}

	private void getPanoramaColorInterpolated(Ray ray) {
		if (skymap == null) {
			scene.sun().skylight(ray);
		} else {
			double r = ray.d.z * ray.d.z + ray.d.x * ray.d.x;
			double theta = 0;
			if (r > Ray.EPSILON)
				theta = Math.asin(ray.d.z / Math.sqrt(r));
			if (ray.d.x < 0)
				theta = Math.PI - theta;
			theta += rotation;
			if (theta > 2 * Math.PI || theta < 0) {
				theta = theta % (2 * Math.PI);
				if (theta < 0)
					theta += 2 * Math.PI;
			}
			double phi = Math.abs(Math.asin(ray.d.y));
			theta /= 2 * Math.PI;
			phi /= Math.PI / 2;
			phi = 1 - phi;
			skymap.getColorInterpolated(theta, phi, ray.color);
		}
	}

	public void getSkyColor4(Ray ray, boolean blackBelowHorizon) {
		if (scene.sun().intersect(ray)) {
			double r = ray.color.x;
			double g = ray.color.y;
			double b = ray.color.z;
			getSkyPanoramaColor(ray, blackBelowHorizon);
			ray.color.x = ray.color.x + r;
			ray.color.y = ray.color.y + g;
			ray.color.z = ray.color.z + b;
			ray.hit = true;
			
		} else {
			getSkyPanoramaColor(ray, blackBelowHorizon);
		}
	}

	public void setRotation(double value) {
		rotation = value;
		scene.refresh();
	}
	
	public double getRotation() {
		return rotation;
	}

	/**
	 * Load sky description from world tag
	 * @param worldTag
	 */
	public void load(CompoundTag worldTag) {
		AnyTag skymapNameTag = worldTag.get("skymapFileName");
		if (!skymapNameTag.isError() && !skymapNameTag.stringValue().isEmpty())
			loadSkyMap(skymapNameTag.stringValue());
		AnyTag rotationTag = worldTag.get("skyYaw");
		if (!rotationTag.isError()) {
			rotation = rotationTag.doubleValue();;
		}
	}

	/**
	 * Save sky description to world tag
	 * @param worldTag
	 */
	public void save(CompoundTag worldTag) {
		if (skymap != null)
			worldTag.addItem("skymapFileName", new StringTag(skymapFileName));
		worldTag.addItem("skyYaw", new DoubleTag(rotation));
	}
}
