/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.renderer.ui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.FileDialog;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;
import java.io.FilenameFilter;
import java.text.NumberFormat;
import java.text.ParseException;

import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.JSlider;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.LayoutStyle;
import javax.swing.SwingUtilities;
import javax.swing.GroupLayout.Alignment;
import javax.swing.LayoutStyle.ComponentPlacement;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

import se.llbit.chunky.main.CenteredFileDialog;
import se.llbit.chunky.main.Chunky;
import se.llbit.chunky.renderer.Camera;
import se.llbit.chunky.renderer.RenderManager;
import se.llbit.chunky.renderer.RenderStateListener;
import se.llbit.chunky.renderer.Scene;
import se.llbit.chunky.renderer.SceneManager;
import se.llbit.util.ProgramProperties;

@SuppressWarnings("serial")
public class RenderControls extends JDialog {

	private final RenderManager controller;
	private final SceneManager sceneManager;
	
	/**
	 * Number format for current locale.
	 */
	private NumberFormat numberFormat =
			NumberFormat.getInstance();
	
	private JSlider sunYawSlider;
	private JSlider skyRotation;
	private JSlider sunPitchSlider;
	private JSlider focalOffsetSlider;
	private JSlider fovSlider;
	private JSlider gammaSlider;
	private JButton loadSkymapBtn;
	private JTextField widthField;
	private JTextField heightField;
	private JSlider dofSlider;
	private JTextField fovField;
	private JTextField gammaField;
	private JTextField dofField;
	private JTextField focalOffsetField;
	private JButton startRenderBtn;
	private JCheckBox enableEmitters;
	private JCheckBox directLight;
	private JButton saveSceneBtn;
	private JButton loadSceneBtn;
	private JTextArea statusText;
	private JButton saveFrameBtn;
	private JCheckBox stillWaterCB;
	private JTextField sceneNameField;
	private JTextField rayDepthField;
	private JSlider rayDepthSlider;
	private JLabel sceneNameLbl;

	private JCheckBox biomeColorsCB;

	private JButton stopRenderBtn;

	private JCheckBox clearWaterCB;
	private JSlider sunIntensitySlider;
	private JSlider emitterIntensitySlider;

	public RenderControls(Chunky parent, final RenderManager renderController,
			final Chunk3DView view) {
		super(parent.getFrame());
		
		this.controller = renderController;
		sceneManager = new SceneManager(this, renderController);
		sceneManager.start();
		
		setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
		setModalityType(Dialog.ModalityType.MODELESS);
		
		addWindowListener(new WindowListener() {
			@Override
			public void windowOpened(WindowEvent e) {
			}
			@Override
			public void windowIconified(WindowEvent e) {
			}
			@Override
			public void windowDeiconified(WindowEvent e) {
			}
			@Override
			public void windowDeactivated(WindowEvent e) {
			}
			@Override
			public void windowClosing(WindowEvent e) {
				sceneManager.interrupt();
				RenderControls.this.dispose();
			}
			@Override
			public void windowClosed(WindowEvent e) {
				if (!view.isClosed()) {
					WindowEvent closeEvent = new WindowEvent(view, WindowEvent.WINDOW_CLOSING);
					Toolkit.getDefaultToolkit().getSystemEventQueue().postEvent(closeEvent);
				}
			}
			@Override
			public void windowActivated(WindowEvent e) {
			}
		});
		
		updateTitle();
		
		JTabbedPane tabbedPane = new JTabbedPane();
		
		tabbedPane.addTab("General", buildGeneralPane());
		tabbedPane.addTab("Lighting", buildLightingPane());
		tabbedPane.addTab("Sky", buildSkyPane());
		tabbedPane.addTab("Camera", buildCameraPane());
		tabbedPane.addTab("Advanced", buildAdvancedPane());
		
		JLabel renderLbl = new JLabel("Render: ");
		
		startRenderBtn = new JButton("START");
		startRenderBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				if (!renderController.scene().pathTrace()) {
					renderController.scene().startRender();
					startRenderBtn.setText("PAUSE");
					startRenderBtn.repaint();
				} else {
					if (renderController.scene().isPaused()) {
						renderController.scene().resumeRender();
						startRenderBtn.setText("PAUSE");
						startRenderBtn.repaint();
					} else {
						renderController.scene().pauseRender();
						startRenderBtn.setText("RESUME");
						startRenderBtn.repaint();
					}
				}
				stopRenderBtn.setVisible(true);
			}
		});
		
		stopRenderBtn = new JButton("HALT");
		stopRenderBtn.setToolTipText("<html>Warning: this will discard the " +
				"current rendered image!<br>Make sure to save your image " +
				"before stopping the renderer!");
		stopRenderBtn.setForeground(Color.red);
		stopRenderBtn.setVisible(false);
		stopRenderBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				renderController.scene().haltRender();
				startRenderBtn.setText("START");
				startRenderBtn.repaint();
				stopRenderBtn.setVisible(false);
			}
		});
		
		renderController.addRenderStateListener(new RenderStateListener() {
			@Override
			public void renderStateChanged(boolean pathTrace, boolean paused) {
				if (pathTrace) {
					if (paused) {
						startRenderBtn.setText("RESUME");
					} else {
						startRenderBtn.setText("PAUSE");
					}
					stopRenderBtn.setVisible(true);
				} else {
					startRenderBtn.setText("START");
					stopRenderBtn.setVisible(false);
				}
			}
		});
		
		saveFrameBtn = new JButton("Save Current Frame");
		saveFrameBtn.addActionListener(saveFrameListener);
		
		statusText = new JTextArea(7, 30);
		statusText.setEditable(false);
		JScrollPane statusTextScrollPane = new JScrollPane(statusText);
		
		JPanel panel = new JPanel();
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup()
						.addComponent(tabbedPane)
						.addComponent(renderLbl)
						.addGroup(layout.createSequentialGroup()
							.addComponent(startRenderBtn)
							.addPreferredGap(ComponentPlacement.UNRELATED)
							.addComponent(stopRenderBtn))
						.addComponent(saveFrameBtn)
						.addComponent(statusTextScrollPane))
					.addContainerGap());
		layout.setVerticalGroup(
			layout.createSequentialGroup()
				.addContainerGap()
				.addComponent(tabbedPane)
				.addPreferredGap(ComponentPlacement.UNRELATED)
				.addComponent(renderLbl)
				.addPreferredGap(ComponentPlacement.RELATED)
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(startRenderBtn)
					.addComponent(stopRenderBtn))
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addComponent(saveFrameBtn)
				.addPreferredGap(ComponentPlacement.UNRELATED)
				.addComponent(statusTextScrollPane)
				.addContainerGap());
		getContentPane().add(panel);
		
		pack();
	}

	private Component buildAdvancedPane() {
		JLabel rayDepthLbl = new JLabel("Ray depth: ");
		rayDepthField = new JTextField(5);
		rayDepthField.addActionListener(rayDepthFieldListener);
		updateRayDepthField();
		rayDepthSlider = new JSlider(1, 10);
		rayDepthSlider.addChangeListener(rayDepthListener);
		updateRayDepthSlider();
		
		JPanel panel = new JPanel();
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup()
						.addGroup(layout.createSequentialGroup()
							.addComponent(rayDepthLbl)
							.addComponent(rayDepthSlider)
							.addComponent(rayDepthField)))
					.addContainerGap());
		layout.setVerticalGroup(
			layout.createSequentialGroup()
				.addContainerGap()
				.addGroup(layout.createParallelGroup()
					.addComponent(rayDepthLbl)
					.addComponent(rayDepthSlider)
					.addComponent(rayDepthField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addContainerGap());
		return panel;
	}

	private Component buildGeneralPane() {
		JLabel widthLbl = new JLabel("Canvas width: ");
		JLabel heightLbl = new JLabel("Canvas height: ");
		JLabel gammaLbl = new JLabel("gamma correction: ");
		
		widthField = new JTextField(10);
		widthField.addActionListener(canvasSizeListener);
		heightField = new JTextField(10);
		heightField.addActionListener(canvasSizeListener);
		
		updateWidthField();
		updateHeightField();
		
		gammaField = new JTextField(5);
		gammaField.addActionListener(gammaFieldListener);
		updateGammaField();
		
		saveSceneBtn = new JButton("Save Scene");
		saveSceneBtn.addActionListener(saveSceneListener);
		
		loadSceneBtn = new JButton("Load Scene");
		loadSceneBtn.addActionListener(loadSceneListener);
		
		loadSceneBtn.setToolTipText("This replaces the current scene!");
		JButton setCanvasSizeBtn = new JButton("Set Canvas Size");
		setCanvasSizeBtn.addActionListener(canvasSizeListener);
		JButton halfCanvasSizeBtn = new JButton("Halve");
		halfCanvasSizeBtn.setToolTipText("Halve the canvas size");
		halfCanvasSizeBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				try {
					controller.setHalveCanvasSize();
					widthField.setText("" + controller.getCanvasWidth());
					heightField.setText("" + controller.getCanvasHeight());
				} catch (InterruptedException e1) {
				}
			}
		});
		JButton doubleCanvasSizeBtn = new JButton("Double");
		halfCanvasSizeBtn.setToolTipText("Double the canvas size");
		doubleCanvasSizeBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				try {
					controller.setDoubleCanvasSize();
					widthField.setText("" + controller.getCanvasWidth());
					heightField.setText("" + controller.getCanvasHeight());
				} catch (InterruptedException e1) {
				}
			}
		});
		
		gammaSlider = new JSlider(1, 100);
		gammaSlider.addChangeListener(gammaListener);
		updateGammaSlider();
		
		JSeparator sep1 = new JSeparator();
		JSeparator sep2 = new JSeparator();
		
		sceneNameLbl = new JLabel("Scene name: ");
		sceneNameField = new JTextField(15); 
		sceneNameField.setText(controller.scene().getSceneName());
		sceneNameField.getDocument().addDocumentListener(new DocumentListener() {
			@Override
			public void removeUpdate(DocumentEvent e) {
				updateName(e);
			}
			@Override
			public void insertUpdate(DocumentEvent e) {
				updateName(e);
			}
			@Override
			public void changedUpdate(DocumentEvent e) {
				updateName(e);
			}
			private void updateName(DocumentEvent e) {
				try {
					Document d = e.getDocument();
					controller.scene().setName(d.getText(0, d.getLength()));
					updateTitle();
				} catch (BadLocationException e1) {
					e1.printStackTrace();
				}
			}
		});
		
		stillWaterCB = new JCheckBox("still water");
		stillWaterCB.addActionListener(stillWaterListener);
		updateStillWater();
		
		clearWaterCB = new JCheckBox("clear water");
		stillWaterCB.addActionListener(clearWaterListener);
		updateClearWater();
		
		biomeColorsCB = new JCheckBox("enable biome colors");
		updateBiomeColorsCB();
		
		JCheckBox postProcessCB = new JCheckBox("enable gamma correction");
		postProcessCB.setSelected(controller.scene().getGammaCorrectionEnabled());
		postProcessCB.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				JCheckBox source = (JCheckBox) e.getSource();
				controller.scene().setGammaCorrectionEnabled(source.isSelected());
			}
		});
		
		final JCheckBox saveDumps = new JCheckBox("save dump once every ");
		saveDumps.setSelected(true);
		JLabel dumpFrequencyLbl = new JLabel(" frame");
		final int[] frequencies = { 50, 100, 500, 1000 };
		String[] frequencyStrings = new String[frequencies.length];
		for (int i = 0; i < frequencies.length; ++i)
			frequencyStrings[i] = frequencies[i] + "th";
		final JComboBox dumpFrequency = new JComboBox(frequencyStrings);
		dumpFrequency.setSelectedIndex(2);
		dumpFrequency.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				int index = dumpFrequency.getSelectedIndex();
				index = Math.max(0, index);
				index = Math.min(frequencies.length, index);
				controller.dumpInterval = frequencies[index];
			}
		});
		saveDumps.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				controller.saveDumps = saveDumps.isSelected();
				dumpFrequency.setEnabled(controller.saveDumps);
			}
		});
		
		JPanel panel = new JPanel();
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup()
						.addGroup(layout.createSequentialGroup()
							.addComponent(sceneNameLbl)
							.addComponent(sceneNameField))
						.addComponent(saveSceneBtn)
						.addComponent(loadSceneBtn)
						.addComponent(sep1)
						.addGroup(layout.createSequentialGroup()
							.addGroup(layout.createParallelGroup()
								.addComponent(widthLbl)
								.addComponent(heightLbl))
							.addGroup(layout.createParallelGroup()
								.addComponent(widthField)
								.addComponent(heightField)))
						.addGroup(layout.createSequentialGroup()
							.addComponent(setCanvasSizeBtn)
							.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
							.addComponent(halfCanvasSizeBtn)
							.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
							.addComponent(doubleCanvasSizeBtn))
						.addComponent(sep2)
						.addGroup(layout.createSequentialGroup()
							.addComponent(gammaLbl)
							.addComponent(gammaSlider)
							.addComponent(gammaField))
						.addComponent(postProcessCB)
						.addComponent(stillWaterCB)
						.addComponent(clearWaterCB)
						.addComponent(biomeColorsCB)
						.addGroup(layout.createSequentialGroup()
							.addComponent(saveDumps)
							.addComponent(dumpFrequency, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE)
							.addComponent(dumpFrequencyLbl)
							.addGap(0, 0, Short.MAX_VALUE)))
					.addContainerGap());
		layout.setVerticalGroup(
			layout.createSequentialGroup()
				.addContainerGap()
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(sceneNameLbl)
					.addComponent(sceneNameField))
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addComponent(saveSceneBtn)
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addComponent(loadSceneBtn)
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addComponent(sep1)
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(widthLbl)
					.addComponent(widthField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(heightLbl)
					.addComponent(heightField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addGroup(layout.createParallelGroup()
					.addComponent(setCanvasSizeBtn)
					.addComponent(halfCanvasSizeBtn)
					.addComponent(doubleCanvasSizeBtn))
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addComponent(sep2)
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addGroup(layout.createParallelGroup()
					.addComponent(gammaLbl)
					.addComponent(gammaSlider)
					.addComponent(gammaField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addComponent(postProcessCB)
				.addComponent(stillWaterCB)
				.addComponent(clearWaterCB)
				.addComponent(biomeColorsCB)
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(saveDumps)
					.addComponent(dumpFrequency)
					.addComponent(dumpFrequencyLbl))
				.addContainerGap());
		return panel;
	}
	
	private Component buildLightingPane() {
		
		final JColorChooser colorChooser = new JColorChooser(Color.white);
		
		colorChooser.setPreviewPanel(new JPanel());
		colorChooser.getSelectionModel().addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent arg0) {
				controller.scene().sun().setColor(colorChooser.getColor());
			}
		});
		
		directLight = new JCheckBox("enable sunlight");
		directLight.setSelected(controller.scene().getDirectLight());
		directLight.addActionListener(directLightListener);
		
		enableEmitters = new JCheckBox("enable emitters");
		enableEmitters.setSelected(controller.scene().getEmittersEnabled());
		enableEmitters.addActionListener(emittersListener);
		
		JLabel sunYawLbl = new JLabel("Sun yaw: ");
		JLabel sunPitchLbl = new JLabel("Sun pitch: ");

		JLabel emitterIntensityLbl = new JLabel("Emitter intensity: ");
		emitterIntensitySlider = new JSlider(1, 100);
		emitterIntensitySlider.addChangeListener(emitterIntensityListener);
		updateEmitterIntensitySlider();
		
		JLabel sunIntensityLbl = new JLabel("Sun intensity: ");
		sunIntensitySlider = new JSlider(0, 100);
		sunIntensitySlider.addChangeListener(sunIntensityListener);
		updateSunIntensitySlider();
		
		sunYawSlider = new JSlider(0, 100);
		sunYawSlider.addChangeListener(sunYawListener);
		
		sunPitchSlider = new JSlider(0, 100);
		sunPitchSlider.addChangeListener(sunPitchListener);
		
		updateSunYawSlider();
		updateSunPitchSlider();
		
		JPanel panel = new JPanel();
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup()
						.addComponent(directLight)
						.addComponent(enableEmitters)
						.addGroup(layout.createSequentialGroup()
							.addGroup(layout.createParallelGroup()
								.addComponent(emitterIntensityLbl)
								.addComponent(sunIntensityLbl)
								.addComponent(sunYawLbl)
								.addComponent(sunPitchLbl))
							.addGroup(layout.createParallelGroup()
								.addComponent(emitterIntensitySlider)
								.addComponent(sunIntensitySlider)
								.addComponent(sunYawSlider)
								.addComponent(sunPitchSlider)))
						.addComponent(colorChooser)));
		layout.setVerticalGroup(
			layout.createSequentialGroup()
				.addContainerGap()
				.addComponent(enableEmitters)
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addGroup(layout.createParallelGroup()
					.addComponent(emitterIntensityLbl)
					.addComponent(emitterIntensitySlider))
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addComponent(directLight)
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addGroup(layout.createParallelGroup()
					.addComponent(sunIntensityLbl)
					.addComponent(sunIntensitySlider))
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addGroup(layout.createParallelGroup()
					.addComponent(sunYawLbl)
					.addComponent(sunYawSlider))
				.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
				.addGroup(layout.createParallelGroup()
					.addComponent(sunPitchLbl)
					.addComponent(sunPitchSlider))
				.addPreferredGap(LayoutStyle.ComponentPlacement.UNRELATED)
				.addComponent(colorChooser)
				.addContainerGap());
		return panel;
	}
	
	private Component buildSkyPane() {
		
		JLabel skyRotationLbl = new JLabel("Skymap rotation:");
		skyRotation = new JSlider(0, 100);
		skyRotation.addChangeListener(skyRotationListener);
		updateSkyRotation();
		
		loadSkymapBtn = new JButton("Load Skymap");
		loadSkymapBtn.setToolTipText("Use a panoramic skymap");
		loadSkymapBtn.addActionListener(loadSkymapListener);
		
		JButton unloadSkymapBtn = new JButton("Unload Skymap");
		unloadSkymapBtn.setToolTipText("Use the default sky");
		unloadSkymapBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				controller.scene().sky().unloadSkymap();
			}
		});
		
		JPanel panel = new JPanel();
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup()
						.addGroup(layout.createSequentialGroup()
							.addComponent(skyRotationLbl)
							.addComponent(skyRotation))
						.addGroup(layout.createSequentialGroup()
							.addComponent(loadSkymapBtn)
							.addPreferredGap(ComponentPlacement.RELATED)
							.addComponent(unloadSkymapBtn)))
					.addContainerGap());
		layout.setVerticalGroup(
			layout.createSequentialGroup()
				.addContainerGap()
				.addGroup(layout.createParallelGroup()
					.addComponent(loadSkymapBtn)
					.addComponent(unloadSkymapBtn))
				.addPreferredGap(ComponentPlacement.UNRELATED)
				.addGroup(layout.createParallelGroup(Alignment.BASELINE)
					.addComponent(skyRotationLbl)
					.addComponent(skyRotation))
				.addContainerGap());
		return panel;
	}
	
	private Component buildCameraPane() {
		
		JLabel fovLbl = new JLabel("Field of View (zoom): ");
		JLabel dofLbl = new JLabel("Depth of Field: ");
		JLabel focalOffsetLbl = new JLabel("Focal Offset: ");

		dofSlider = new JSlider(1, 1000);
		dofSlider.addChangeListener(dofListener);
		updateDofSlider();
		
		fovSlider = new JSlider(1, 1000);
		fovSlider.addChangeListener(fovListener);
		updateFovSlider();
		
		focalOffsetSlider = new JSlider(1, 1000);
		focalOffsetSlider.addChangeListener(focalOffsetListener);
		updateFocalOffsetSlider();
		
		fovField = new JTextField(5);
		fovField.addActionListener(fovFieldListener);
		updateFovField();
		
		dofField = new JTextField(5);
		dofField.addActionListener(dofFieldListener);
		updateDofField();
		
		focalOffsetField = new JTextField(5);
		focalOffsetField.addActionListener(focalOffsetFieldListener);
		updateFocalOffsetField();
		
		JButton autoFocusBtn = new JButton("Autofocus");
		autoFocusBtn.setToolTipText("Focuses on the object right in the center, under the crosshairs");
		autoFocusBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				controller.scene().autoFocus();
				updateDofField();
				updateDofSlider();
				updateFocalOffsetField();
				updateFocalOffsetSlider();
			}
		});
		
		JButton cameraToPlayerBtn = new JButton("Camera to player");
		cameraToPlayerBtn.setToolTipText("Move camera to player position");
		cameraToPlayerBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				controller.scene().moveCameraToPlayer();
			}
		});
		
		JButton centerCameraBtn = new JButton("Center camera");
		centerCameraBtn.setToolTipText("Center camera above loaded chunks");
		centerCameraBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				controller.scene().centerCamera();
			}
		});
		
		JSeparator sep1 = new JSeparator();
		
		JPanel panel = new JPanel();
		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createSequentialGroup()
					.addContainerGap()
					.addGroup(layout.createParallelGroup()
						.addComponent(cameraToPlayerBtn)
						.addComponent(centerCameraBtn)
						.addComponent(sep1)
						.addGroup(layout.createSequentialGroup()
							.addGroup(layout.createParallelGroup()
								.addComponent(fovLbl)
								.addComponent(dofLbl)
								.addComponent(focalOffsetLbl))
							.addGroup(layout.createParallelGroup()
								.addComponent(fovSlider)
								.addComponent(dofSlider)
								.addComponent(focalOffsetSlider))
							.addGroup(layout.createParallelGroup()
								.addComponent(fovField)
								.addComponent(dofField)
								.addComponent(focalOffsetField)))
						.addComponent(autoFocusBtn))
					.addContainerGap());
		layout.setVerticalGroup(
			layout.createSequentialGroup()
				.addContainerGap()
				.addComponent(cameraToPlayerBtn)
				.addPreferredGap(ComponentPlacement.RELATED)
				.addComponent(centerCameraBtn)
				.addPreferredGap(ComponentPlacement.UNRELATED)
				.addComponent(sep1)
				.addPreferredGap(ComponentPlacement.UNRELATED)
				.addGroup(layout.createParallelGroup()
					.addComponent(fovLbl)
					.addComponent(fovSlider)
					.addComponent(fovField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addGroup(layout.createParallelGroup()
					.addComponent(dofLbl)
					.addComponent(dofSlider)
					.addComponent(dofField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addGroup(layout.createParallelGroup()
					.addComponent(focalOffsetLbl)
					.addComponent(focalOffsetSlider)
					.addComponent(focalOffsetField, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
				.addPreferredGap(ComponentPlacement.UNRELATED)
				.addComponent(autoFocusBtn)
				.addContainerGap());
		return panel;
	}

	protected void updateStillWater() {
		stillWaterCB.removeActionListener(stillWaterListener);
		stillWaterCB.setSelected(controller.scene().stillWaterEnabled());
		stillWaterCB.addActionListener(stillWaterListener);
	}
	
	protected void updateClearWater() {
		clearWaterCB.removeActionListener(clearWaterListener);
		clearWaterCB.setSelected(controller.scene().getClearWater());
		clearWaterCB.addActionListener(clearWaterListener);
	}
	
	protected void updateBiomeColorsCB() {
		biomeColorsCB.removeActionListener(biomeColorsCBListener);
		biomeColorsCB.setSelected(controller.scene().biomeColorsEnabled());
		biomeColorsCB.addActionListener(biomeColorsCBListener);
	}

	private void updateTitle() {
		setTitle("Render Controls - " + controller.scene().getSceneName());
	}

	ActionListener canvasSizeListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			try {
				controller.setCanvasSize(
						Integer.parseInt(widthField.getText()),
						Integer.parseInt(heightField.getText()));
				widthField.setText("" + controller.getCanvasWidth());
				heightField.setText("" + controller.getCanvasHeight());
				ProgramProperties.setProperty("3dcanvas.width", "" + controller.getCanvasWidth());
				ProgramProperties.setProperty("3dcanvas.height", "" + controller.getCanvasHeight());
			} catch (NumberFormatException e1) {
				Chunky.logger.info("Failed to set canvas size: invalid dimensions!");
			} catch (InterruptedException e1) {
			}
		}
	};
	ActionListener saveSceneListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			sceneManager.saveScene(sceneNameField.getText());
		}
	};
	ActionListener saveFrameListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			controller.saveFrame();
		}
	};
	ActionListener loadSceneListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			new SceneSelector(RenderControls.this);
		}
	};
	ActionListener loadSkymapListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			CenteredFileDialog fileDialog =
					new CenteredFileDialog(null, "Open Skymap", FileDialog.LOAD);
			fileDialog.setDirectory(System.getProperty("user.dir"));
			fileDialog.setFilenameFilter(
					new FilenameFilter() {
						@Override
						public boolean accept(File dir, String name) {
							return name.toLowerCase().endsWith(".png")
									|| name.toLowerCase().endsWith(".jpg");
						}
					});
			fileDialog.setVisible(true);
			File selectedFile = fileDialog.getSelectedFile();
			if (selectedFile != null) {
				controller.scene().sky().loadSkyMap(selectedFile.getAbsolutePath());
			}
		}
	};
	ChangeListener rayDepthListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			controller.scene().setRayDepth(source.getValue());
			updateRayDepthField();
		}
	};
	ActionListener rayDepthFieldListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JTextField source = (JTextField) e.getSource();
			try {
				int value = Integer.parseInt(source.getText());
				controller.scene().setRayDepth(Math.max(1, value));
				updateRayDepthSlider();
			} catch (NumberFormatException ex) {
			}
		}
	};
	ChangeListener emitterIntensityListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			controller.scene().setEmitterIntensity(source.getValue()*2);
		}
	};
	ChangeListener sunIntensityListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			controller.scene().sun().setIntensity(source.getValue() / 25.);
		}
	};
	ChangeListener sunYawListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			controller.scene().sun().setSunYaw(source.getValue() * Math.PI / 50);
		}
	};
	ChangeListener sunPitchListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			controller.scene().sun().setSunPitch(source.getValue() * Math.PI / 200);
		}
	};
	ChangeListener dofListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			if (source.getValue() == source.getMaximum()) {
				controller.scene().camera().setInfDof(true);
			} else {
				controller.scene().camera().setInfDof(false);
				double value = (double) (source.getValue() - source.getMinimum())
						/ (source.getMaximum() - source.getMinimum());
				double scale = Camera.MAX_DOF - Camera.MIN_DOF;
				controller.scene().camera().setDof(value * scale + Camera.MIN_DOF);
			}
			updateDofField();
		}
	};
	ChangeListener skyRotationListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			double value = (double) (source.getValue() - source.getMinimum())
					/ (source.getMaximum() - source.getMinimum());
			controller.scene().sky().setRotation(value * 2 * Math.PI);
			updateDofField();
		}
	};
	ChangeListener fovListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			double value = (double) (source.getValue() - source.getMinimum())
					/ (source.getMaximum() - source.getMinimum());
			double scale = Camera.MAX_FOV - Camera.MIN_FOV;
			controller.scene().camera().setFov(value * scale + Camera.MIN_FOV);
			updateFovField();
		}
	};
	ChangeListener gammaListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			double value = (double) (source.getValue() - source.getMinimum())
					/ (source.getMaximum() - source.getMinimum());
			double scale = Scene.MAX_GAMMA
				- Scene.MIN_GAMMA;
			controller.scene().setGamma(
					value * scale + Scene.MIN_GAMMA);
			updateGammaField();
		}
	};
	ChangeListener focalOffsetListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider) e.getSource();
			double value = (double) (source.getValue() - source.getMinimum())
					/ (source.getMaximum() - source.getMinimum());
			double scale = Camera.MAX_FOCAL_OFFSET - Camera.MIN_FOCAL_OFFSET;
			controller.scene().camera().setFocalOffset(value * scale + Camera.MIN_FOCAL_OFFSET);
			updateFocalOffsetField();
		}
	};
	ActionListener stillWaterListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			controller.scene().setStillWater(stillWaterCB.isSelected());
		}
	};
	ActionListener clearWaterListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			controller.scene().setClearWater(clearWaterCB.isSelected());
		}
	};
	ActionListener biomeColorsCBListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			controller.scene().setBiomeColorsEnabled(biomeColorsCB.isSelected());
		}
	};
	ActionListener emittersListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			controller.scene().setEmittersEnabled(enableEmitters.isSelected());
		}
	};
	ActionListener directLightListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			controller.scene().setDirectLight(directLight.isSelected());
		}
	};
	ActionListener fovFieldListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JTextField source = (JTextField) e.getSource();
			try {
				double value = numberFormat.parse(source.getText()).doubleValue();
				value = Math.max(value, Camera.MIN_FOV);
				value = Math.min(value, Camera.MAX_FOV);
				controller.scene().camera().setFov(value);
				updateFovSlider();
			} catch (NumberFormatException ex) {
			} catch (ParseException ex) {
			}
		}
	};
	ActionListener gammaFieldListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JTextField source = (JTextField) e.getSource();
			try {
				double value = numberFormat.parse(source.getText()).doubleValue();
				value = Math.max(value, Scene.MIN_GAMMA);
				value = Math.min(value, Scene.MAX_GAMMA);
				controller.scene().setGamma(value);
				updateGammaSlider();
			} catch (NumberFormatException ex) {
			} catch (ParseException ex) {
			}
		}
	};
	ActionListener dofFieldListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JTextField source = (JTextField) e.getSource();
			if (source.getText().equals("inf")) {
				controller.scene().camera().setInfDof(true);
			} else {
				try {
					double value = numberFormat.parse(source.getText()).doubleValue();
					value = Math.max(value, Camera.MIN_DOF);
					controller.scene().camera().setInfDof(false);
					controller.scene().camera().setDof(value);
					updateDofSlider();
				} catch (NumberFormatException ex) {
				} catch (ParseException ex) {
				}
			}
		}
	};
	ActionListener focalOffsetFieldListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JTextField source = (JTextField) e.getSource();
			try {
				double value = numberFormat.parse(source.getText()).doubleValue();
				value = Math.max(value, Camera.MIN_FOCAL_OFFSET);
				controller.scene().camera().setFocalOffset(value);
				updateFocalOffsetSlider();
			} catch (NumberFormatException ex) {
			} catch (ParseException ex) {
			}
		}
	};
	
	protected void updateRayDepthSlider() {
		rayDepthSlider.removeChangeListener(rayDepthListener);
		rayDepthSlider.setValue(controller.scene().getRayDepth());
		rayDepthSlider.addChangeListener(rayDepthListener);
	}
	
	protected void updateRayDepthField() {
		rayDepthField.removeActionListener(rayDepthFieldListener);
		rayDepthField.setText("" + controller.scene().getRayDepth());
		rayDepthField.addActionListener(rayDepthFieldListener);
	}
	
	protected void updateDofField() {
		dofField.removeActionListener(dofFieldListener);
		if (controller.scene().camera().getInfDof())
			dofField.setText("inf");
		else
			dofField.setText(String.format("%.2f", controller.scene().camera().getDof()));
		dofField.addActionListener(dofFieldListener);
	}
	
	protected void updateFocalOffsetField() {
		focalOffsetField.removeActionListener(focalOffsetFieldListener);
		focalOffsetField.setText(String.format("%.2f", controller.scene().camera().getFocalOffset()));
		focalOffsetField.addActionListener(focalOffsetFieldListener);
	}

	protected void updateFovField() {
		fovField.removeActionListener(fovFieldListener);
		fovField.setText(String.format("%.2f", controller.scene().camera().getFov()));
		fovField.addActionListener(fovFieldListener);
	}
	
	protected void updateSkyRotation() {
		skyRotation.removeChangeListener(skyRotationListener);
		skyRotation.setValue((int) Math.round(
				100 * controller.scene().sky().getRotation() / (2 * Math.PI)));
		skyRotation.addChangeListener(skyRotationListener);
	}
	
	protected void updateGammaField() {
		gammaField.removeActionListener(gammaFieldListener);
		gammaField.setText(String.format("%.2f", controller.scene().getGamma()));
		gammaField.addActionListener(gammaFieldListener);
	}
	
	protected void updateSunPitchSlider() {
		sunPitchSlider.removeChangeListener(sunPitchListener);
		sunPitchSlider.setValue((int) (controller.scene().sun().getSunPitch() * 200 / Math.PI));
		sunPitchSlider.addChangeListener(sunPitchListener);
	}
	
	protected void updateEmitterIntensitySlider() {
		emitterIntensitySlider.removeChangeListener(emitterIntensityListener);
		emitterIntensitySlider.setValue((int) (controller.scene().getEmitterIntensity()/2));
		emitterIntensitySlider.addChangeListener(emitterIntensityListener);
	}
	
	protected void updateSunIntensitySlider() {
		sunIntensitySlider.removeChangeListener(sunIntensityListener);
		sunIntensitySlider.setValue((int) (controller.scene().sun().getIntensity() * 25));
		sunIntensitySlider.addChangeListener(sunIntensityListener);
	}
	
	protected void updateSunYawSlider() {
		sunYawSlider.removeChangeListener(sunYawListener);
		sunYawSlider.setValue((int) (controller.scene().sun().getSunYaw() * 50 / Math.PI));
		sunYawSlider.addChangeListener(sunYawListener);
	}
	
	protected void updateDofSlider() {
		dofSlider.removeChangeListener(dofListener);
		if (controller.scene().camera().getInfDof()) {
			dofSlider.setValue(dofSlider.getMaximum());
		} else {
			double value = (controller.scene().camera().getDof() - Camera.MIN_DOF)
					/ (Camera.MAX_DOF - Camera.MIN_DOF);
			double scale = dofSlider.getMaximum() - dofSlider.getMinimum();
			dofSlider.setValue((int) (value * scale + dofSlider.getMinimum()));
		}
		dofSlider.addChangeListener(dofListener);
	}
	
	protected void updateFocalOffsetSlider() {
		focalOffsetSlider.removeChangeListener(focalOffsetListener);
		double value = (controller.scene().camera().getFocalOffset() - Camera.MIN_FOCAL_OFFSET)
				/ (Camera.MAX_FOCAL_OFFSET - Camera.MIN_FOCAL_OFFSET);
		double scale = focalOffsetSlider.getMaximum() - focalOffsetSlider.getMinimum();
		focalOffsetSlider.setValue((int) (value * scale + focalOffsetSlider.getMinimum()));
		focalOffsetSlider.addChangeListener(focalOffsetListener);
	}

	protected void updateFovSlider() {
		fovSlider.removeChangeListener(fovListener);
		double value = (controller.scene().camera().getFov() - Camera.MIN_FOV)
				/ (Camera.MAX_FOV - Camera.MIN_FOV);
		double scale = fovSlider.getMaximum() - fovSlider.getMinimum();
		fovSlider.setValue((int) (value * scale + fovSlider.getMinimum()));
		fovSlider.addChangeListener(fovListener);
	}

	protected void updateGammaSlider() {
		gammaSlider.removeChangeListener(gammaListener);
		double value = (controller.scene().getGamma()
				- Scene.MIN_GAMMA)
				/ (Scene.MAX_GAMMA - Scene.MIN_GAMMA);
		double scale = gammaSlider.getMaximum() - gammaSlider.getMinimum();
		gammaSlider.setValue((int) (value * scale + gammaSlider.getMinimum()));
		gammaSlider.addChangeListener(gammaListener);
	}

	public void setStatusText(final String text) {
		SwingUtilities.invokeLater(new Runnable() {
			@Override
			public void run() {
				statusText.setText(text);
			}
		});
	}
	
	protected void updateWidthField() {
		widthField.setText("" + controller.getCanvasWidth());
	}
	
	protected void updateHeightField() {
		heightField.setText("" + controller.getCanvasHeight());
	}

	public void loadScene(String sceneName) {
		sceneManager.loadScene(sceneName);
	}
	
	public void sceneDescriptionLoaded() {
		updateDofField();
		updateFovField();
		updateGammaField();
		updateFocalOffsetField();
		updateDofSlider();
		updateFovSlider();
		updateGammaSlider();
		updateFocalOffsetSlider();
		updateWidthField();
		updateHeightField();
		updateEmitterIntensitySlider();
		updateSunIntensitySlider();
		updateSunYawSlider();
		updateSunPitchSlider();
		updateStillWater();
		updateClearWater();
		updateSkyRotation();
		updateBiomeColorsCB();
		updateTitle();
		sceneNameField.setText(controller.scene().getSceneName());
		enableEmitters.setSelected(controller.scene().getEmittersEnabled());
		directLight.setSelected(controller.scene().getDirectLight());
		startRenderBtn.setText("RESUME");
		stopRenderBtn.setVisible(true);
	}
	
	public void sceneSaved() {
		updateTitle();
	}
}
